#!/usr/bin/env python3
"""

    Add placename data to iceaddr database.

    Fetches placename data from the IS50V geospatial database and inserts it
    into the iceaddr database. Also adds manual placename additions from
    placename_additions.txt.

"""

from typing import List, Tuple

from pprint import pprint  # type: ignore
import sqlite3
from pathlib import Path
from io import BytesIO
import zipfile

import fiona  # type: ignore
import requests

from iceaddr.dist import in_iceland


ORNEFNI_DATA_FILE = "is_50v_ornefni_wgs_84.gpkg"
ORNEFNI_DATA_URL = (
    "https://atlas.lmi.is/heikir/downloadData/is_50v_ornefni_wgs_84_gpkg.zip"
)

# Remote URL for latest IS50V data:
# https://atlas.lmi.is/heikir/downloadData/is_50v_ornefni_wgs_84_gpkg.zip
# This file should be placed in repo root and renamed before running this program
GPKG_FILE = "ornefni.gpkg"

LAYERS = [
    "ornefni_flakar",
    "ornefni_linur",
    "ornefni_punktar",
]


DEFAULT_DBNAME = "iceaddr.db"


def fetch_ornefni_data() -> None:
    """Fetch IS50V placename data from remote URL, unzip
    from memory to current directory and rename file."""

    r = requests.get(ORNEFNI_DATA_URL, allow_redirects=True)
    if r.status_code != 200:
        print(f"Failed to download {ORNEFNI_DATA_URL}")
        exit(1)

    z = zipfile.ZipFile(BytesIO(r.content))
    z.extractall()

    Path(ORNEFNI_DATA_FILE).rename(GPKG_FILE)


def center_point(coords: List[Tuple[float, float]]) -> Tuple[float, float]:
    """Find the center point of a given set of coordinates."""
    x: float = 0
    y: float = 0

    for px, py in coords:
        x += px
        y += py

    x = x / len(coords)
    y = y / len(coords)

    return (x, y)


def create_table(dbpath: str) -> sqlite3.Connection:
    """Create ornefni database table."""
    dbconn = sqlite3.connect(dbpath)

    create_table_sql = """
    CREATE TABLE ornefni (
        id INTEGER UNIQUE PRIMARY KEY NOT NULL,
        nafn TEXT,
        flokkur TEXT,
        lat_wgs84 REAL,
        long_wgs84 REAL
    );
    """

    try:
        dbconn.cursor().execute(create_table_sql)
    except Exception:
        print("Unable to create table 'ornefni'")
        exit()

    return dbconn


def delete_table(dbpath: str) -> sqlite3.Connection:
    """Drop ornefni database table."""
    dbconn = sqlite3.connect(dbpath)

    del_table_sql = """DROP TABLE ornefni"""

    try:
        dbconn.cursor().execute(del_table_sql)
        print("Deleted pre-existing table 'ornefni'")
    except Exception:
        pass

    return dbconn


def add_placename_additions(dbc: sqlite3.Connection) -> None:
    """Read manual placename additions from text file, insert into "ornefni" DB table."""
    print("Inserting placename additions")
    f = open("placename_additions.txt", "r")
    for line in f.readlines():
        if not line.strip() or line.strip().startswith("#"):
            continue
        comp = line.split(":")
        first = comp[0]
        last = comp[-1].strip()
        try:
            (latstr, lonstr, fl) = last.split(",")
            lat = float(latstr) if latstr else None
            lon = float(lonstr) if lonstr else None
        except Exception as e:
            print(f"{line}: error: {e}")
            raise

        # print("Inserting " + first)

        if lat and lon and not in_iceland((lat, lon)):
            raise Exception(f"Not in Iceland: {first} ({lat}, {lon})")

        dbc.cursor().execute(
            "INSERT INTO ornefni (nafn, flokkur, lat_wgs84, long_wgs84) VALUES (?,?,?,?)",
            (first, fl, lat, lon),
        )
    f.close()
    dbc.commit()


def add_placenames_from_is50v(dbc: sqlite3.Connection) -> None:
    """Read IS50V geo layers from file, add all placenames ("örnefni") to DB."""
    if not Path(GPKG_FILE).exists():
        print(f"Could not find file {GPKG_FILE}")
        exit(1)

    for layer in fiona.listlayers(GPKG_FILE):
        with fiona.open(GPKG_FILE, encoding="utf-8", layer=layer) as src:
            for i in src:
                wanted_layer = [layer.startswith(p) for p in LAYERS]
                if True not in wanted_layer:
                    # print("Skipping layer " + layer)
                    continue

                # pprint(i)
                try:
                    # pprint(i["properties"])
                    fl = i["properties"]["ornefnaflokkur"]
                    n = i["properties"]["ornefni"]
                    c = i["geometry"]["coordinates"]
                    # nc = len(c)
                except Exception as e:
                    print(f"ERROR adding item {i['properties']}: {e}")
                else:
                    pass
                    # print(n)

                # Special handling of lines (e.g. rivers)
                if layer.startswith("ornefni_linur"):
                    firstcoords = c[0]

                    if type(firstcoords[0]) is list:
                        firstcoords = firstcoords[0]

                    if type(firstcoords) is list:
                        firstcoords = firstcoords[0]
                    cp = firstcoords

                # Special handling of flakes - use center point
                elif type(c) is list:
                    if not c:
                        # print(f"Faulty flake: {n}")
                        # pprint(i)
                        # raise
                        continue

                    firstcoords = c[0]
                    t = type(firstcoords[0])
                    if t is float:
                        cp = firstcoords
                    elif t is list:
                        firstcoords = firstcoords[0]
                        cp = center_point(firstcoords)

                # Just a point
                else:
                    cp = c

                # LMÍ's GPKG coord values are reversed! Why?
                gps = (cp[1], cp[0])

                if not in_iceland(gps):
                    print(f"WARNING: Not in Iceland, skipping: {n} ({gps})")

                # Insert
                dbc.cursor().execute(
                    "INSERT INTO ornefni (nafn, flokkur, lat_wgs84, long_wgs84) VALUES (?,?,?,?)",
                    (n, fl, gps[0], gps[1]),
                )

            dbc.commit()


def main() -> None:
    # Fetch placename data from remote URL
    fetch_ornefni_data()

    # Delete any existing table
    delete_table(DEFAULT_DBNAME)

    # Start by creating ornefni table
    dbc = create_table(DEFAULT_DBNAME)
    if not dbc:
        print("Failed to connect to DB")
        exit(1)

    add_placename_additions(dbc)
    add_placenames_from_is50v(dbc)


if __name__ == "__main__":
    """Command line invocation."""
    main()



#!/usr/bin/env python3
"""

    Create stadfong address database from from the
    Staðfangaskrá CSV dataset.

    From data compiled by Registers Iceland (CC-BY):
        https://opingogn.is/dataset/stadfangaskra
        https://opendefinition.org/licenses/cc-by/

"""

from typing import Iterator, Dict, Any

from builtins import input

import sys
import os
import sqlite3
import csv
from pathlib import Path
from io import BytesIO, TextIOWrapper
from urllib.request import urlopen
from iceaddr.dist import in_iceland

import humanize  # type: ignore


STADFONG_REMOTE_URL = "https://fasteignaskra.is/Stadfangaskra.csv"

DSV_FILENAME = "Stadfangaskra.csv"

DEFAULT_DBNAME = "iceaddr.db"

COLS = [
    "hnitnum",
    "svfnr",
    "byggd",
    "landnr",
    "postnr",
    "heiti_nf",
    "heiti_tgf",
    "husnr",
    "bokst",
    "serheiti",
    "vidsk",
    "lat_wgs84",
    "long_wgs84",
]


def create_db(path: str) -> sqlite3.Connection:
    """Create stadfong database table."""
    dbconn = sqlite3.connect(path)

    create_table_sql = """
    CREATE TABLE stadfong (
        hnitnum INTEGER UNIQUE PRIMARY KEY NOT NULL,
        svfnr INTEGER,
        byggd INTEGER,
        landnr INTEGER,
        postnr INTEGER,
        heiti_nf TEXT,
        heiti_tgf TEXT,
        husnr INTEGER,
        bokst TEXT,
        serheiti TEXT,
        vidsk TEXT,
        lat_wgs84 REAL,
        long_wgs84 REAL
    );
    """

    dbconn.cursor().execute(create_table_sql)

    return dbconn


def read_rows(
    dsv_file: TextIOWrapper, delimiter: str = ","
) -> Iterator[Dict[Any, Any]]:
    reader = csv.DictReader(dsv_file, delimiter=delimiter)
    for row in reader:
        yield row


def insert_address_entry(e: Dict[Any, Any], conn: sqlite3.Connection) -> None:
    # The stadfong datafile is quite dirty so we need to
    # sanitise values before inserting into the database

    for k in e:
        e[k] = e[k].strip()

    # Icelandic to English decimal points
    e["LAT_WGS84"] = e["N_HNIT_WGS84"].replace(",", ".")
    e["LONG_WGS84"] = e["E_HNIT_WGS84"].replace(",", ".")

    to_int = ["BYGGD", "HEINUM", "HNITNUM", "HUSNR", "LANDNR", "POSTNR"]
    to_float = ["LAT_WGS84", "LONG_WGS84"]

    for k in to_int:
        if e[k] == "":
            e[k] = None
        else:
            try:
                e[k] = int(e[k])
            except Exception:
                print("Failed to convert '" + k + "' to int, setting to null")
                e[k] = None

    for k in to_float:
        if e[k] == "":
            e[k] = None
        else:
            try:
                e[k] = float(e[k])
            except Exception:
                print("Failed to convert '" + k + "' to float, setting to null")
                e[k] = None

    try:
        assert in_iceland((e["LAT_WGS84"], e["LONG_WGS84"]))
        qargs = [e[c.upper()] for c in COLS]
        # print(qargs)
        c = conn.cursor()
        c.execute("INSERT INTO stadfong VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?)", qargs)
    except Exception as exc:
        print(exc)


def main() -> None:
    # Optional args to specify input and output files
    stadfong_path = sys.argv[1] if len(sys.argv) > 1 else DSV_FILENAME
    db_path = sys.argv[2] if len(sys.argv) > 2 else DEFAULT_DBNAME

    if stadfong_path == DSV_FILENAME and not Path(stadfong_path).is_file():
        print("Fetching remote file %s" % STADFONG_REMOTE_URL)
        resp = urlopen(STADFONG_REMOTE_URL)
        f = TextIOWrapper(BytesIO(resp.read()), "utf-8")
    else:
        f = open(stadfong_path, "r")

    # Delete previous db file
    if Path(db_path).is_file():
        if input("%s exists, overwrite? (y/n): " % db_path).lower().startswith("y"):
            os.remove(db_path)
        else:
            print("Aborting")
            sys.exit()

    # Create new database file
    dbconn = create_db(db_path)

    # Commit to DB in chunks of 1000 rows
    cnt = 0
    for r in read_rows(f):
        insert_address_entry(r, dbconn)
        cnt += 1
        if cnt % 1000 == 0:
            dbconn.commit()
            print("\tInserting: %d\r" % cnt, end="")
            sys.stdout.flush()

    dbconn.commit()

    print("\tInserting: %d\r" % cnt, end="")
    sys.stdout.flush()

    bytesize: int = os.stat(db_path).st_size
    human_size = humanize.naturalsize(bytesize)  # type: ignore

    print("\nCreated database with %d entries (%s)" % (cnt, human_size))


if __name__ == "__main__":
    """Command line invocation."""
    main()



"""

    iceaddr: Look up information about Icelandic streets, addresses,
             placenames, landmarks, locations and postcodes.

    This file contains code related to distance calculation.

"""

from __future__ import annotations

from typing import Tuple

import math

_EARTH_RADIUS = 6371.0088  # Earth's radius in km


def distance(loc1: Tuple[float, float], loc2: Tuple[float, float]) -> float:
    """
    Calculate the Haversine distance.
    Parameters
    ----------
    origin : tuple of float
        (lat, long)
    destination : tuple of float
        (lat, long)
    Returns
    -------
    distance_in_km : float
    Examples
    --------
    >>> origin = (48.1372, 11.5756)  # Munich
    >>> destination = (52.5186, 13.4083)  # Berlin
    >>> round(distance(origin, destination), 1)
    504.2
    Source:
    https://stackoverflow.com/questions/19412462
        /getting-distance-between-two-points-based-on-latitude-longitude
    """
    (lat1, lon1) = loc1
    (lat2, lon2) = loc2

    # Bad params, or missing coordinates, return infinity for distance-sorting purposes
    if not lat1 or not lon1 or not lat2 or not lon2:
        return float("inf")

    dlat = math.radians(lat2 - lat1)
    dlon = math.radians(lon2 - lon1)
    slat = math.sin(dlat / 2)
    slon = math.sin(dlon / 2)
    a = (
        slat * slat
        + math.cos(math.radians(lat1)) * math.cos(math.radians(lat2)) * slon * slon
    )
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    return _EARTH_RADIUS * c


ICELAND_COORDS = (64.9957538607, -18.5739616708)


def in_iceland(loc: Tuple[float, float], km_radius: float = 800.0) -> bool:
    """Check if coordinates are within or very close to Iceland."""
    return distance(loc, ICELAND_COORDS) <= km_radius



"""

    iceaddr: Look up information about Icelandic streets, addresses,
             placenames, landmarks, locations and postcodes.

    Copyright (c) 2018-2023 Sveinbjorn Thordarson.

    This file contains code related to placename lookup.

"""

from __future__ import annotations

from typing import Any, Dict, List

from .db import shared_db
from .dist import distance

# These particular placenames share a name with other, perhaps larger, placenames,
# but should never the less be given priority when ordering results.
HARDCODED_PRIORITY = {
    "Hellisheiði": (64.0221268, -21.3413149),  # Nálægt Rvk fær forgang
    "Snæfellsnes": (64.8731746, -23.0309911),  # Nesið norðan við Reykjanes!
    "Mýrdalur": (63.4462885, -19.0832988),  # Nálægt Vík
    "Mosfellsheiði": (64.1675067, -21.3733656),  # Nálægt Mosó
    "Bláfjöll": (64.0121886, -21.5617119),  # Nálægt Rvk á Reykjanesskaga
    "Bakki": (66.0701681, -17.3481556),  # Hjá Húsavík, sbr. verið
    "Bessastaðir": (64.1059036227962, -21.9957549156328),  # Forsetabústaður
    "Gullfoss": (64.3273264, -20.1193949),  # Túrista-áfangastaðurinn fær forgang
    "Grótta": (64.1642163, -22.0218824),  # Á Seltjarnarnesi fær forgang
    "Arnarhóll": (64.147844, -21.9331656),  # Arnarhóll í miðborg Rvk
    "Reykjanes": (63.8185821975681, -22.692991355433815),  # Nesið nálægt Rvk.
}

# This determines the sort order of results
# if there's more than one placename match.
ORDER = [
    "Dummy",  # Index 0 is reserved for hardcoded priority
    "Sveitarfélag",
    "Þéttbýli",
    "Sveit",
    "Sýsla",
    "Hreppur",
    "Flugvöllur",
    "Jarðgöng",
    "Virkjun",
    "Kirkja",
    "Landörnefni Stórt",
    "Jökla- og snævarörnefni Stórt",
    "Sjávarörnefni Stórt",
    "Vatnaörnefni Stórt",
    "Landörnefni Mið",
    "Jökla- og snævarörnefni Mið",
    "Sjávarörnefni Mið",
    "Vatnaörnefni Mið",
    "Landörnefni Lítið",
    "Jökla- og snævarörnefni Lítið",
    "Sjávarörnefni Lítið",
    "Vatnaörnefni Lítið",
]


def _precedence(pn: Dict[str, Any]) -> int:
    """Sort priority for placenames."""
    if pn["nafn"] in HARDCODED_PRIORITY:
        (lat, lng) = HARDCODED_PRIORITY[pn["nafn"]]
        if pn["lat_wgs84"] == lat and pn["long_wgs84"] == lng:
            return 0

    fl = pn["flokkur"]
    if fl in ORDER:
        return ORDER.index(fl)
    # Any number > len(ORDER) will do for sorting purposes
    return 9999


def placename_lookup(placename: str, partial: bool = False) -> List[Dict[str, Any]]:
    """Look up Icelandic placename in database."""
    q = "SELECT * FROM ornefni WHERE nafn=?"
    if partial:
        q = "SELECT * FROM ornefni WHERE nafn LIKE '?%'"

    db_conn = shared_db.connection()
    res = db_conn.cursor().execute(q, [placename])
    matches = [dict(row) for row in res]
    matches.sort(key=_precedence)

    return matches


def nearest_placenames(lat: float, lon: float, limit: int = 1) -> List[Dict[str, Any]]:
    """Find the placename closest to the given coordinates."""
    q = "SELECT * FROM ornefni"
    db_conn = shared_db.connection()
    res = db_conn.cursor().execute(q, [])
    closest = sorted(
        res, key=lambda i: distance((lat, lon), (i["lat_wgs84"], i["long_wgs84"]))
    )
    return closest[:limit]



"""

    iceaddr: Look up information about Icelandic streets, addresses,
             placenames, landmarks, locations and postcodes.

    Copyright (c) 2018-2023 Sveinbjorn Thordarson.

"""

__author__ = "Sveinbjorn Thordarson"
__version__ = "0.5.7"  # Also update setup.py

from .addresses import *  # noqa: F403
from .placenames import *  # noqa: F403
from .postcodes import *  # noqa: F403



#!/usr/bin/env python3
"""

    Fetch list of postcodes from the Icelandic postal service,
    create and print dict mapping postcode to placename and
    other related information. Also adds placenames in the
    nominative case (nefnifall) since the source data only
    includes placenames in dative (þágufall).

    https://www.postur.is/gogn/Gotuskra/postnumer.txt

"""

import requests
import pprint
import logging
import csv
from io import StringIO

from reynir import NounPhrase

from iceaddr.postcodes import POSTCODES


POSTCODES_REMOTE_URL = "https://www.postur.is/gogn/Gotuskra/postnumer.txt"


def _clean_name(name: str) -> str:
    return name.split(" - ")[0].strip()


def main() -> None:
    pc = dict(POSTCODES)
    pc_keys = pc.keys()
    pp = pprint.PrettyPrinter(indent=4)

    req = requests.get(POSTCODES_REMOTE_URL, allow_redirects=True)
    f = StringIO(req.text)

    changed = False
    reader = csv.DictReader(f, delimiter=";")
    for r in reader:
        # CSV file from postur.is only contains postcode placenames in
        # the dative form (þgf.). Try to lemmatise to nominative (nf.) using Reynir.
        postcode = int(r["Póstnúmer"])
        if postcode not in pc_keys:
            logging.warning(f"Postcode '{postcode}' did not already exist in data.")
            changed = True

        tp = r["Tegund"]
        p_dat = _clean_name(r["Staður"])
        p_nom = NounPhrase(p_dat).nominative
        if not p_nom:
            logging.warning(f"Unable to decline placename '{p_dat}'")
            p_nom = p_dat

        if pc[postcode]["stadur_nf"] != p_nom:
            pc[postcode]["stadur_nf"] = p_nom
            print(f"{pc[postcode]['stadur_nf']} --> {p_nom}")
            changed = True

        if pc[postcode]["stadur_tgf"] != p_dat:
            pc[postcode]["stadur_tgf"] = p_dat
            print(f"{pc[postcode]['stadur_tgf']} --> {p_dat}")
            changed = True

        if pc[postcode]["tegund"] != tp:
            pc[postcode]["tegund"] = tp
            print(f"{pc[postcode]['tegund']} --> {tp}")
            changed = True

    if not changed:
        print("No change since last update")
    else:
        pp.pprint(pc)


if __name__ == "__main__":
    """Command line invocation."""
    main()



"""

    iceaddr: Look up information about Icelandic streets, addresses,
             placenames, landmarks, locations and postcodes.

    Copyright (c) 2018-2023 Sveinbjorn Thordarson.

    This file contains code related to the iceaddr sqlite3 database.

"""

import sqlite3

from pkg_resources import resource_filename

_DB_REL_PATH = "iceaddr.db"


class SharedDB:
    """Singleton object wrapper for local SQLite3 database."""

    def __init__(self):
        self.db_conn = None

    def connection(self) -> sqlite3.Connection:
        # Open connection lazily
        if not self.db_conn:
            db_path = resource_filename(__name__, _DB_REL_PATH)

            # Open database file in read-only mode via URI
            db_uri = f"file:{db_path}?mode=ro"
            self.db_conn = sqlite3.connect(db_uri, uri=True, check_same_thread=False)

            # Return rows as key-value dicts
            self.db_conn.row_factory = lambda c, r: dict(  # type: ignore noqa: PGH003
                zip([col[0] for col in c.description], r)  # type: ignore noqa: PGH003
            )

        return self.db_conn


shared_db = SharedDB()



"""

    iceaddr: Look up information about Icelandic streets, addresses,
             placenames, landmarks, locations and postcodes.

    Copyright (c) 2018-2023 Sveinbjorn Thordarson.

    Tests for iceaddr python package.

"""

from typing import Dict, Optional

import sys
import os

# Add parent directory to import path
sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)) + "/../")

from iceaddr import (  # noqa
    iceaddr_lookup,
    iceaddr_suggest,
    postcode_lookup,
    postcodes_for_placename,
    postcodes_for_region,
    POSTCODES,
    nearest_addr,
    nearest_placenames,
)


def test_address_lookup():
    """Test address lookup using various known addresses."""
    ADDR_TO_POSTCODE = [
        ("Öldugata", 4, "Reykjavík", 101),
        ("öldugötu", 12, "hafnarfirði", 220),
        ("Tómasarhaga", 12, "Reykjavík", 107),
        ("smiðjuvegur", 22, "", 200),
    ]

    for a in ADDR_TO_POSTCODE:
        res = iceaddr_lookup(a[0], number=a[1], placename=a[2])
        assert res and res[0]["postnr"] == a[3]

    res = iceaddr_lookup("Brattagata", number=4, letter="b")
    assert res and res[0]["postnr"] == 310 and res[0]["stadur_nf"] == "Borgarnes"

    POSTCODE_TO_PLACENAME = [
        ("Öldugata", 4, 101, "Reykjavík", "Höfuðborgarsvæðið", "Þéttbýli"),
        (
            "dagverðardalur",
            11,
            400,
            "Ísafjörður",
            "Vesturland og Vestfirðir",
            "Þéttbýli",
        ),
        ("Höfðabraut", 3, 805, "Selfoss", "Suðurland og Reykjanes", "Stærra dreifbýli"),
    ]

    for p in POSTCODE_TO_PLACENAME:
        # print(f"iceaddr_lookup('{p[0]}', number={p[1]}, postcode={p[2]}"
        res = iceaddr_lookup(p[0], number=p[1], postcode=p[2])
        assert res[0]["stadur_nf"] == p[3]
        assert res[0]["svaedi_nf"] == p[4]
        # assert res[0]["tegund"] == p[5]

    assert len(iceaddr_lookup("Stærri-Árskógi", postcode=621)) > 0
    assert len(iceaddr_lookup("Grundarstíg")) > 10
    assert len(iceaddr_lookup("Grundarstíg", limit=2)) == 2


def test_address_lookup_matches_number_range_and_no_number():
    results = iceaddr_lookup("Vesturgata", number=6, placename="Reykjavík")
    assert results
    assert results[0]["vidsk"] == "6-8"
    assert results[0]["husnr"] is None


# def test_address_lookup_can_find_places_of_interest():
#     results = iceaddr_lookup("Harpa", postcode=101)
#     assert results
#     assert results[0]["heiti_nf"] == "Austurbakki"
#     assert results[0]["husnr"]


def test_address_lookup_does_not_need_letter():
    assert iceaddr_lookup("Laugavegur", number=151)
    assert iceaddr_lookup("Laugavegur", number=151, letter="r")
    assert not iceaddr_lookup("Laugavegur", number=151, letter="e")


def test_address_suggestions():
    """Test address suggestions for natural language search strings."""
    res = iceaddr_suggest("Öldugötu 4, 101")
    assert res[0]["heiti_nf"] == "Öldugata"
    assert res[0]["husnr"] == 4
    assert res[0]["stadur_nf"] == "Reykjavík"

    res = iceaddr_suggest("Öldugata 4, Rey")
    assert [n["stadur_tgf"] for n in res] == ["Reykjavík", "Reyðarfirði"]

    res = iceaddr_suggest("öldugötu 4b, 621")
    assert res[0]["bokst"].lower() == "b"
    assert res[0]["stadur_tgf"] == "Dalvík"

    res = iceaddr_suggest("öldugötu 4B, 621")
    assert res and res[0]["bokst"].lower() == "b"

    assert iceaddr_suggest("Öldugata a4B") == []
    assert iceaddr_suggest("Öldugötu 4Baaa") == []

    assert len(iceaddr_suggest("Kl")) == 0  # always empty for fewer than 3 chars
    assert len(iceaddr_suggest("öldu", limit=75)) == 75


def test_address_suggest_with_dashed_numbers():
    assert iceaddr_suggest("Laugavegur 151-155")


def test_postcode_data_integrity():
    """Make sure postcode data is sane."""
    for k, v in POSTCODES.items():
        assert type(k) == int
        _verify_postcode_dict(v)


def _verify_postcode_dict(pcd: Optional[Dict[str, str]]):
    """Verify the integrity of a postcode dict."""
    assert pcd is not None
    assert "svaedi_nf" in pcd
    assert "svaedi_tgf" in pcd
    assert "stadur_nf" in pcd
    assert "stadur_tgf" in pcd
    assert "tegund" in pcd
    assert "lysing" not in pcd or pcd["lysing"] != ""


def test_postcode_lookup():
    """Test postcode lookup functions."""
    _verify_postcode_dict(postcode_lookup(101))
    _verify_postcode_dict(postcode_lookup(900))
    _record102 = postcode_lookup("102")
    _verify_postcode_dict(_record102)
    assert _record102 and _record102["stadur_nf"] == "Reykjavík"

    kop_pc_num = 5
    assert len(postcodes_for_placename("Kópavogur")) == kop_pc_num
    assert len(postcodes_for_placename("kópavogi")) == kop_pc_num
    assert len(postcodes_for_placename("kópav", partial=True)) == kop_pc_num

    selfoss_pc = [800, 801, 802, 803, 804, 805, 806]
    assert postcodes_for_placename("Selfossi") == selfoss_pc
    assert postcodes_for_placename("selfoss") == selfoss_pc
    assert postcodes_for_placename("SELFOS", partial=True) == selfoss_pc

    assert postcodes_for_region("Norðurland")
    assert postcodes_for_region("Höfuðborgarsvæðið")


FISKISLOD_31_COORDS = (64.1560233, -21.951407)
OLDUGATA_4_COORDS = (64.148446, -21.944933)


def test_nearest_addr():
    """Test address proxmity function."""
    addr = nearest_addr(FISKISLOD_31_COORDS[0], FISKISLOD_31_COORDS[1])
    assert len(addr) == 1
    assert addr[0]["heiti_nf"] == "Fiskislóð"
    assert addr[0]["postnr"] == 101
    assert addr[0]["svaedi_nf"] == "Höfuðborgarsvæðið"

    addr = nearest_addr(OLDUGATA_4_COORDS[0], OLDUGATA_4_COORDS[1], limit=3)
    assert len(addr) == 3
    assert addr[0]["heiti_nf"] == "Öldugata"
    assert addr[0]["postnr"] == 101
    assert addr[0]["svaedi_tgf"] == "Höfuðborgarsvæðinu"


def test_nearest_placename():
    """Test placename proximity function."""
    pn = nearest_placenames(FISKISLOD_31_COORDS[0], FISKISLOD_31_COORDS[1])
    assert len(pn) == 1
    assert pn[0]["nafn"] == "Grandi"

    pn = nearest_placenames(OLDUGATA_4_COORDS[0], OLDUGATA_4_COORDS[1], limit=5)
    assert len(pn) == 5
    assert "Landakotshæð" in [x["nafn"] for x in pn]



"""

    iceaddr: Look up information about Icelandic streets, addresses,
             placenames, landmarks, locations and postcodes.

    Copyright (c) 2018-2023 Sveinbjorn Thordarson.

    This file contains code and data related to Icelandic postcodes.

"""

from __future__ import annotations

from typing import Dict, List, Optional, Union

POSTCODES: Dict[int, Dict[str, str]] = {
    101: {
        "lysing": "Miðborg",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    102: {
        "lysing": "Vatnsmýri og Skerjafjörður",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    103: {
        "lysing": "Háaleitis- og Bústaðahverfi",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    104: {
        "lysing": "Laugardalur",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    105: {
        "lysing": "Hlíðar",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    107: {
        "lysing": "Vesturbær",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    108: {
        "lysing": "Háaleitis- og Bústaðahverfi",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    109: {
        "lysing": "Breiðholt",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    110: {
        "lysing": "Árbær",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    111: {
        "lysing": "Breiðholt",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    112: {
        "lysing": "Grafarvogur",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    113: {
        "lysing": "Grafarholt og Úlfarsárdalur",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    116: {
        "lysing": "Kjalarnes",
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    121: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    123: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    124: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    125: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    127: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    128: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    129: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    130: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    132: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    150: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Dreifbýli",
    },
    155: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Dreifbýli",
    },
    161: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Dreifbýli",
    },
    162: {
        "stadur_nf": "Reykjavík",
        "stadur_tgf": "Reykjavík",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Dreifbýli",
    },
    170: {
        "stadur_nf": "Seltjarnarnes",
        "stadur_tgf": "Seltjarnarnesi",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    172: {
        "stadur_nf": "Seltjarnarnes",
        "stadur_tgf": "Seltjarnarnesi",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    190: {
        "stadur_nf": "Vogar",
        "stadur_tgf": "Vogum",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    191: {
        "stadur_nf": "Vogar",
        "stadur_tgf": "Vogum",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    200: {
        "stadur_nf": "Kópavogur",
        "stadur_tgf": "Kópavogi",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    201: {
        "stadur_nf": "Kópavogur",
        "stadur_tgf": "Kópavogi",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    202: {
        "stadur_nf": "Kópavogur",
        "stadur_tgf": "Kópavogi",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    203: {
        "stadur_nf": "Kópavogur",
        "stadur_tgf": "Kópavogi",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    206: {
        "stadur_nf": "Kópavogur",
        "stadur_tgf": "Kópavogi",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Dreifbýli",
    },
    210: {
        "stadur_nf": "Garðabær",
        "stadur_tgf": "Garðabæ",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    212: {
        "stadur_nf": "Garðabær",
        "stadur_tgf": "Garðabæ",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    220: {
        "stadur_nf": "Hafnarfjörður",
        "stadur_tgf": "Hafnarfirði",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    221: {
        "stadur_nf": "Hafnarfjörður",
        "stadur_tgf": "Hafnarfirði",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    222: {
        "stadur_nf": "Hafnarfjörður",
        "stadur_tgf": "Hafnarfirði",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Pósthólf",
    },
    225: {
        "stadur_nf": "Garðabær",
        "stadur_tgf": "Garðabæ",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    230: {
        "stadur_nf": "Reykjanesbær",
        "stadur_tgf": "Reykjanesbæ",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    232: {
        "stadur_nf": "Reykjanesbær",
        "stadur_tgf": "Reykjanesbæ",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Pósthólf",
    },
    233: {
        "stadur_nf": "Reykjanesbær",
        "stadur_tgf": "Reykjanesbæ",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    235: {
        "lysing": "Keflavíkurflugvöllur",
        "stadur_nf": "Keflavíkurflugvöllur",
        "stadur_tgf": "Keflavíkurflugvelli",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    240: {
        "stadur_nf": "Grindavík",
        "stadur_tgf": "Grindavík",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    241: {
        "stadur_nf": "Grindavík",
        "stadur_tgf": "Grindavík",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    245: {
        "stadur_nf": "Suðurnesjabær",
        "stadur_tgf": "Suðurnesjabæ",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    246: {
        "stadur_nf": "Suðurnesjabær",
        "stadur_tgf": "Suðurnesjabæ",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    250: {
        "stadur_nf": "Suðurnesjabær",
        "stadur_tgf": "Suðurnesjabæ",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    251: {
        "stadur_nf": "Suðurnesjabær",
        "stadur_tgf": "Suðurnesjabæ",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    260: {
        "stadur_nf": "Reykjanesbær",
        "stadur_tgf": "Reykjanesbæ",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    262: {
        "lysing": "Ásbrú",
        "stadur_nf": "Reykjanesbær",
        "stadur_tgf": "Reykjanesbæ",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    270: {
        "stadur_nf": "Mosfellsbær",
        "stadur_tgf": "Mosfellsbæ",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    271: {
        "stadur_nf": "Mosfellsbær",
        "stadur_tgf": "Mosfellsbæ",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    276: {
        "lysing": "Kjós",
        "stadur_nf": "Mosfellsbær",
        "stadur_tgf": "Mosfellsbæ",
        "svaedi_nf": "Höfuðborgarsvæðið",
        "svaedi_tgf": "Höfuðborgarsvæðinu",
        "tegund": "Stærra dreifbýli",
    },
    300: {
        "stadur_nf": "Akranes",
        "stadur_tgf": "Akranesi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    301: {
        "stadur_nf": "Akranes",
        "stadur_tgf": "Akranesi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    302: {
        "stadur_nf": "Akranes",
        "stadur_tgf": "Akranesi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Pósthólf",
    },
    310: {
        "stadur_nf": "Borgarnes",
        "stadur_tgf": "Borgarnesi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    311: {
        "stadur_nf": "Borgarnes",
        "stadur_tgf": "Borgarnesi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    320: {
        "stadur_nf": "Reykholt",
        "stadur_tgf": "Reykholti",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    340: {
        "stadur_nf": "Stykkishólmur",
        "stadur_tgf": "Stykkishólmi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    341: {
        "stadur_nf": "Stykkishólmur",
        "stadur_tgf": "Stykkishólmi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    342: {
        "stadur_nf": "Stykkishólmur",
        "stadur_tgf": "Stykkishólmi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    345: {
        "stadur_nf": "Flatey",
        "stadur_tgf": "Flatey",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    350: {
        "stadur_nf": "Grundarfjörður",
        "stadur_tgf": "Grundarfirði",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    351: {
        "stadur_nf": "Grundarfjörður",
        "stadur_tgf": "Grundarfirði",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    355: {
        "stadur_nf": "Ólafsvík",
        "stadur_tgf": "Ólafsvík",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    356: {
        "stadur_nf": "Snæfellsbær",
        "stadur_tgf": "Snæfellsbæ",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    360: {
        "stadur_nf": "Hellissandur",
        "stadur_tgf": "Hellissandi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    370: {
        "stadur_nf": "Búðardalur",
        "stadur_tgf": "Búðardal",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    371: {
        "stadur_nf": "Búðardalur",
        "stadur_tgf": "Búðardal",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    380: {
        "stadur_nf": "Reykhólahreppur",
        "stadur_tgf": "Reykhólahreppi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    381: {
        "stadur_nf": "Reykhólahreppur",
        "stadur_tgf": "Reykhólahreppi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    400: {
        "stadur_nf": "Ísafjörður",
        "stadur_tgf": "Ísafirði",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    401: {
        "stadur_nf": "Ísafjörður",
        "stadur_tgf": "Ísafirði",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    410: {
        "stadur_nf": "Hnífsdalur",
        "stadur_tgf": "Hnífsdal",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    415: {
        "stadur_nf": "Bolungarvík",
        "stadur_tgf": "Bolungarvík",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    416: {
        "stadur_nf": "Bolungarvík",
        "stadur_tgf": "Bolungarvík",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    420: {
        "stadur_nf": "Súðavík",
        "stadur_tgf": "Súðavík",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    421: {
        "stadur_nf": "Súðavík",
        "stadur_tgf": "Súðavík",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    425: {
        "stadur_nf": "Flateyri",
        "stadur_tgf": "Flateyri",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    426: {
        "stadur_nf": "Flateyri",
        "stadur_tgf": "Flateyri",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    430: {
        "stadur_nf": "Suðureyri",
        "stadur_tgf": "Suðureyri",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    431: {
        "stadur_nf": "Suðureyri",
        "stadur_tgf": "Suðureyri",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    450: {
        "stadur_nf": "Patreksfjörður",
        "stadur_tgf": "Patreksfirði",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    451: {
        "stadur_nf": "Patreksfjörður",
        "stadur_tgf": "Patreksfirði",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    460: {
        "stadur_nf": "Tálknafjörður",
        "stadur_tgf": "Tálknafirði",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    461: {
        "stadur_nf": "Tálknafjörður",
        "stadur_tgf": "Tálknafirði",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    465: {
        "stadur_nf": "Bíldudalur",
        "stadur_tgf": "Bíldudal",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    466: {
        "stadur_nf": "Bíldudalur",
        "stadur_tgf": "Bíldudal",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    470: {
        "stadur_nf": "Þingeyri",
        "stadur_tgf": "Þingeyri",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    471: {
        "stadur_nf": "Þingeyri",
        "stadur_tgf": "Þingeyri",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    500: {
        "stadur_nf": "Staður",
        "stadur_tgf": "Stað",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    510: {
        "stadur_nf": "Hólmavík",
        "stadur_tgf": "Hólmavík",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    511: {
        "stadur_nf": "Hólmavík",
        "stadur_tgf": "Hólmavík",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    512: {
        "stadur_nf": "Hólmavík",
        "stadur_tgf": "Hólmavík",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    520: {
        "stadur_nf": "Drangsnes",
        "stadur_tgf": "Drangsnesi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Stærra dreifbýli",
    },
    524: {
        "stadur_nf": "Árneshreppur",
        "stadur_tgf": "Árneshreppi",
        "svaedi_nf": "Vesturland og Vestfirðir",
        "svaedi_tgf": "Vesturlandi og Vestfjörðum",
        "tegund": "Dreifbýli",
    },
    530: {
        "stadur_nf": "Hvammstangi",
        "stadur_tgf": "Hvammstanga",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    531: {
        "stadur_nf": "Hvammstangi",
        "stadur_tgf": "Hvammstanga",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    540: {
        "stadur_nf": "Blönduós",
        "stadur_tgf": "Blönduósi",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    541: {
        "stadur_nf": "Blönduós",
        "stadur_tgf": "Blönduósi",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    545: {
        "stadur_nf": "Skagaströnd",
        "stadur_tgf": "Skagaströnd",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    546: {
        "stadur_nf": "Skagaströnd",
        "stadur_tgf": "Skagaströnd",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    550: {
        "stadur_nf": "Sauðárkrókur",
        "stadur_tgf": "Sauðárkróki",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    551: {
        "stadur_nf": "Sauðárkrókur",
        "stadur_tgf": "Sauðárkróki",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    560: {
        "stadur_nf": "Varmahlíð",
        "stadur_tgf": "Varmahlíð",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    561: {
        "stadur_nf": "Varmahlíð",
        "stadur_tgf": "Varmahlíð",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    565: {
        "stadur_nf": "Hofsós",
        "stadur_tgf": "Hofsósi",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    566: {
        "stadur_nf": "Hofsós",
        "stadur_tgf": "Hofsósi",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    570: {
        "stadur_nf": "Fljót",
        "stadur_tgf": "Fljótum",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    580: {
        "stadur_nf": "Siglufjörður",
        "stadur_tgf": "Siglufirði",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    581: {
        "stadur_nf": "Siglufjörður",
        "stadur_tgf": "Siglufirði",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    600: {
        "stadur_nf": "Akureyri",
        "stadur_tgf": "Akureyri",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    601: {
        "stadur_nf": "Akureyri",
        "stadur_tgf": "Akureyri",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    602: {
        "stadur_nf": "Akureyri",
        "stadur_tgf": "Akureyri",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Pósthólf",
    },
    603: {
        "stadur_nf": "Akureyri",
        "stadur_tgf": "Akureyri",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    604: {
        "stadur_nf": "Akureyri",
        "stadur_tgf": "Akureyri",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    605: {
        "stadur_nf": "Akureyri",
        "stadur_tgf": "Akureyri",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    606: {
        "stadur_nf": "Akureyri",
        "stadur_tgf": "Akureyri",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    607: {
        "stadur_nf": "Akureyri",
        "stadur_tgf": "Akureyri",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    610: {
        "stadur_nf": "Grenivík",
        "stadur_tgf": "Grenivík",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    611: {
        "stadur_nf": "Grímsey",
        "stadur_tgf": "Grímsey",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    616: {
        "stadur_nf": "Grenivík",
        "stadur_tgf": "Grenivík",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    620: {
        "stadur_nf": "Dalvík",
        "stadur_tgf": "Dalvík",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    621: {
        "stadur_nf": "Dalvík",
        "stadur_tgf": "Dalvík",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    625: {
        "stadur_nf": "Ólafsfjörður",
        "stadur_tgf": "Ólafsfirði",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    626: {
        "stadur_nf": "Ólafsfjörður",
        "stadur_tgf": "Ólafsfirði",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    630: {
        "stadur_nf": "Hrísey",
        "stadur_tgf": "Hrísey",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    640: {
        "stadur_nf": "Húsavík",
        "stadur_tgf": "Húsavík",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    641: {
        "stadur_nf": "Húsavík",
        "stadur_tgf": "Húsavík",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    645: {
        "stadur_nf": "Fosshóll",
        "stadur_tgf": "Fosshóli",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    650: {
        "stadur_nf": "Laugar",
        "stadur_tgf": "Laugum",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    660: {
        "stadur_nf": "Mývatn",
        "stadur_tgf": "Mývatni",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    670: {
        "stadur_nf": "Kópasker",
        "stadur_tgf": "Kópaskeri",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    671: {
        "stadur_nf": "Kópasker",
        "stadur_tgf": "Kópaskeri",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    675: {
        "stadur_nf": "Raufarhöfn",
        "stadur_tgf": "Raufarhöfn",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    676: {
        "stadur_nf": "Raufarhöfn",
        "stadur_tgf": "Raufarhöfn",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    680: {
        "stadur_nf": "Þórshöfn",
        "stadur_tgf": "Þórshöfn",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    681: {
        "stadur_nf": "Þórshöfn",
        "stadur_tgf": "Þórshöfn",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    685: {
        "stadur_nf": "Bakkafjörður",
        "stadur_tgf": "Bakkafirði",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    686: {
        "stadur_nf": "Bakkafjörður",
        "stadur_tgf": "Bakkafirði",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    690: {
        "stadur_nf": "Vopnafjörður",
        "stadur_tgf": "Vopnafirði",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Stærra dreifbýli",
    },
    691: {
        "stadur_nf": "Vopnafjörður",
        "stadur_tgf": "Vopnafirði",
        "svaedi_nf": "Norðurland",
        "svaedi_tgf": "Norðurlandi",
        "tegund": "Dreifbýli",
    },
    700: {
        "stadur_nf": "Egilsstaðir",
        "stadur_tgf": "Egilsstöðum",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    701: {
        "stadur_nf": "Egilsstaðir",
        "stadur_tgf": "Egilsstöðum",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    710: {
        "stadur_nf": "Seyðisfjörður",
        "stadur_tgf": "Seyðisfirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    711: {
        "stadur_nf": "Seyðisfjörður",
        "stadur_tgf": "Seyðisfirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    715: {
        "stadur_nf": "Mjóifjörður",
        "stadur_tgf": "Mjóafirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    720: {
        "stadur_nf": "Borgarfjörður eystra",
        "stadur_tgf": "Borgarfirði eystra",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    721: {
        "stadur_nf": "Borgarfjörður eystra",
        "stadur_tgf": "Borgarfirði eystra",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    730: {
        "stadur_nf": "Reyðarfjörður",
        "stadur_tgf": "Reyðarfirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    731: {
        "stadur_nf": "Reyðarfjörður",
        "stadur_tgf": "Reyðarfirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    735: {
        "stadur_nf": "Eskifjörður",
        "stadur_tgf": "Eskifirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    736: {
        "stadur_nf": "Eskifjörður",
        "stadur_tgf": "Eskifirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    740: {
        "stadur_nf": "Neskaupstaður",
        "stadur_tgf": "Neskaupstað",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    741: {
        "stadur_nf": "Neskaupstaður",
        "stadur_tgf": "Neskaupstað",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    750: {
        "stadur_nf": "Fáskrúðsfjörður",
        "stadur_tgf": "Fáskrúðsfirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    751: {
        "stadur_nf": "Fáskrúðsfjörður",
        "stadur_tgf": "Fáskrúðsfirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    755: {
        "stadur_nf": "Stöðvarfjörður",
        "stadur_tgf": "Stöðvarfirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    756: {
        "stadur_nf": "Stöðvarfjörður",
        "stadur_tgf": "Stöðvarfirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    760: {
        "stadur_nf": "Breiðdalsvík",
        "stadur_tgf": "Breiðdalsvík",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    761: {
        "stadur_nf": "Breiðdalsvík",
        "stadur_tgf": "Breiðdalsvík",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    765: {
        "stadur_nf": "Djúpivogur",
        "stadur_tgf": "Djúpavogi",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    766: {
        "stadur_nf": "Djúpivogur",
        "stadur_tgf": "Djúpavogi",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    780: {
        "stadur_nf": "Höfn í Hornafirði",
        "stadur_tgf": "Höfn í Hornafirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    781: {
        "stadur_nf": "Höfn í Hornafirði",
        "stadur_tgf": "Höfn í Hornafirði",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Dreifbýli",
    },
    785: {
        "stadur_nf": "Öræfi",
        "stadur_tgf": "Öræfum",
        "svaedi_nf": "Austurland",
        "svaedi_tgf": "Austurlandi",
        "tegund": "Stærra dreifbýli",
    },
    800: {
        "stadur_nf": "Selfoss",
        "stadur_tgf": "Selfossi",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    801: {
        "stadur_nf": "Selfoss",
        "stadur_tgf": "Selfossi",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    802: {
        "stadur_nf": "Selfoss",
        "stadur_tgf": "Selfossi",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Pósthólf",
    },
    803: {
        "stadur_nf": "Selfoss",
        "stadur_tgf": "Selfossi",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    804: {
        "stadur_nf": "Selfoss",
        "stadur_tgf": "Selfossi",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    805: {
        "stadur_nf": "Selfoss",
        "stadur_tgf": "Selfossi",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    806: {
        "stadur_nf": "Selfoss",
        "stadur_tgf": "Selfossi",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    810: {
        "stadur_nf": "Hveragerði",
        "stadur_tgf": "Hveragerði",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    815: {
        "stadur_nf": "Þorlákshöfn",
        "stadur_tgf": "Þorlákshöfn",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    816: {
        "stadur_nf": "Ölfus",
        "stadur_tgf": "Ölfusi",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    820: {
        "stadur_nf": "Eyrarbakki",
        "stadur_tgf": "Eyrarbakka",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    825: {
        "stadur_nf": "Stokkseyri",
        "stadur_tgf": "Stokkseyri",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    840: {
        "stadur_nf": "Laugarvatn",
        "stadur_tgf": "Laugarvatni",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    845: {
        "stadur_nf": "Flúðir",
        "stadur_tgf": "Flúðum",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    846: {
        "stadur_nf": "Flúðir",
        "stadur_tgf": "Flúðum",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    850: {
        "stadur_nf": "Hella",
        "stadur_tgf": "Hellu",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    851: {
        "stadur_nf": "Hella",
        "stadur_tgf": "Hellu",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    860: {
        "stadur_nf": "Hvolsvöllur",
        "stadur_tgf": "Hvolsvelli",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    861: {
        "stadur_nf": "Hvolsvöllur",
        "stadur_tgf": "Hvolsvelli",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    870: {
        "stadur_nf": "Vík",
        "stadur_tgf": "Vík",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    871: {
        "stadur_nf": "Vík",
        "stadur_tgf": "Vík",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    880: {
        "stadur_nf": "Kirkjubæjarklaustur",
        "stadur_tgf": "Kirkjubæjarklaustri",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    881: {
        "stadur_nf": "Kirkjubæjarklaustur",
        "stadur_tgf": "Kirkjubæjarklaustri",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Dreifbýli",
    },
    900: {
        "stadur_nf": "Vestmannaeyjar",
        "stadur_tgf": "Vestmannaeyjum",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Stærra dreifbýli",
    },
    902: {
        "stadur_nf": "Vestmannaeyjar",
        "stadur_tgf": "Vestmannaeyjum",
        "svaedi_nf": "Suðurland og Reykjanes",
        "svaedi_tgf": "Suðurlandi og Reykjanesi",
        "tegund": "Pósthólf",
    },
}


def postcode_lookup(postcode: Union[int, str]) -> Optional[Dict[str, str]]:
    """Return postcode info dictionary given a postcode.
    Accepts either numeric or string argument."""
    return POSTCODES.get(int(postcode))


def postcodes_for_region(region_name: str, partial: bool = False) -> List[int]:
    """Return postcodes matching a full or partial region name,
    e.g. "Norðurland", "Höfuðborgarsvæðið"."""
    return _filter_postcodes("svaedi", region_name, partial=partial)


def postcodes_for_placename(placename: str, partial: bool = False) -> List[int]:
    """Returns postcodes matching a full or partial placename,
    e.g. "Reykjavík", "Dalvík"."""
    return _filter_postcodes("stadur", placename, partial=partial)


def _filter_postcodes(key: str, searchstr: str, partial: bool = False) -> List[int]:
    """Utility function to find postcodes matching a criterion."""
    assert key in ["stadur", "svaedi"]
    p = searchstr.lower()
    matches: List[int] = []

    k1 = key + "_nf"
    k2 = key + "_tgf"

    for k, v in POSTCODES.items():
        nf = v[k1].lower()
        tgf = v[k2].lower()
        if (partial and (nf.startswith(p) or tgf.startswith(p))) or p in (nf, tgf):
            matches.append(k)

    return matches



"""

    iceaddr: Look up information about Icelandic streets, addresses,
             placenames, landmarks, locations and postcodes.

    Copyright (c) 2018-2023 Sveinbjorn Thordarson.

    This file contains code related to Icelandic address lookup.

"""

from __future__ import annotations

from typing import Any, Dict, List, Optional

import re

from .db import shared_db
from .dist import distance
from .postcodes import POSTCODES, postcodes_for_placename


def _add_postcode_info(addr: Dict[str, Any]) -> Dict[str, Any]:
    """Look up postcode info, add keys to address dictionary."""
    pn = addr.get("postnr")
    if pn and POSTCODES.get(pn):
        addr.update(POSTCODES[pn])
    return addr


def _run_addr_query(q: str, qargs: List[str]) -> List[Dict[str, Any]]:
    """Run address query, w. additional postcode data added post hoc."""
    db_conn = shared_db.connection()
    res = db_conn.cursor().execute(q, qargs)
    return [_add_postcode_info(dict(row)) for row in res]


def _cap_first(s: str) -> str:
    """Returns string with first character capitalized. Why this isn't in
    the Python stdlib is beyond me. The capitalize() function annoyingly
    lowercases the rest of the string."""
    return s[:1].upper() + s[1:] if s else s


def iceaddr_lookup(  # noqa: PLR0913
    street_name: str,
    number: Optional[int] = None,
    letter: Optional[str] = None,
    postcode: Optional[int] = None,
    placename: Optional[str] = None,
    limit: int = 50,
) -> List[Dict[str, Any]]:
    """Look up all addresses matching criterion"""

    # Be forgiving, strip and capitalize street name. All street names in DB are capitalized.
    street_name = _cap_first(street_name.strip())

    pc = [postcode] if postcode else []

    # Look up postcodes for placename if no postcode is provided
    if placename and not postcode:
        pc = postcodes_for_placename(placename.strip())
    q = "SELECT * FROM stadfong WHERE"
    name_fields = ["heiti_nf=?", "heiti_tgf=?"]
    if not number:
        # Add lookup for churches and places of interest like Harpa
        name_fields.append("serheiti=?")
    q += "({})".format(" OR ".join(name_fields))
    sqlargs = [street_name] * len(name_fields)

    if number:
        q += " AND (husnr=? OR substr(vidsk, 0, instr(vidsk, '-')) = ?)"
        sqlargs.append(str(number))
        sqlargs.append(str(number))
        if letter:
            q += " AND bokst LIKE ? COLLATE NOCASE"
            sqlargs.append(letter)

    if pc:
        qp = " OR ".join([" postnr=?" for _ in pc])
        sqlargs.extend([str(x) for x in pc])
        q += " AND (%s) " % qp

    # Ordering by postcode may in fact be a reasonable proxy
    # for delivering by order of match likelihood since the
    # lowest postcodes are generally more densely populated
    q += " ORDER BY vidsk != '', postnr ASC, husnr ASC, bokst ASC LIMIT ?"
    sqlargs.append(str(limit))

    return _run_addr_query(q, sqlargs)


MIN_SEARCH_STR_LEN = 3


def iceaddr_suggest(search_str: str, limit: int = 50) -> List[Dict[str, Any]]:
    """Parse search string and fetch matching addresses.
    Made to handle partial and full text queries in
    the following formats:

    Öldug
    Öldugata
    Öldugata 4
    Öldugata 4, 101
    Öldugata 4, Reykjavík
    Öldugata 4, 101 Reykjavík
    """

    search_str = _cap_first(search_str.strip())
    if not search_str or len(search_str) < MIN_SEARCH_STR_LEN:
        return []

    items = [s.strip().split() for s in search_str.split(",")]

    if not [a for a in items if len(a)]:
        return []  # Nothing to search for

    # Street name component
    addr = items[0]

    # Handle street names with more than one word, or trailing character
    # E.g. "Stærri Bær 1", "Bárugata 17a"
    if re.match(r"\d+", addr[-1]):
        addr = [" ".join(addr[:-1]), addr[-1]]
        m = re.search(r"([a-zA-Z])$", addr[-1])
        if m:
            addr[-1] = addr[-1][:-1]
            addr.append(m.group(0).lower())
    else:
        addr = [" ".join(addr)]

    q = "SELECT * FROM stadfong WHERE "
    qargs: List[str] = []

    street_name = addr[0]
    if len(addr) == 1:  # "Ölduga"
        q += " (heiti_nf LIKE ? OR heiti_tgf LIKE ?) "
        qargs.extend([street_name + "%", street_name + "%"])
    elif len(addr) >= 2:  # noqa: PLR2004 "Öldugötu 4"
        # Street name
        q += " (heiti_nf=? OR heiti_tgf=?) "
        qargs.extend([street_name, street_name])

        # Street number
        if "-" in addr[1]:
            # "Viðskeyti við staðfang", this is where dashed number ranges are
            q += " AND vidsk=?"
        else:
            q += " AND husnr=?"
        qargs.append(addr[1])

        # Street number's trailing character
        # e.g. if it's "Öldugata 4b"
        if len(addr) == 3:  # noqa: PLR2004
            q += " AND bokst LIKE ? COLLATE NOCASE"
            qargs.append(addr[2])

    # Placename component (postcode or placename)
    if len(items) > 1 and items[1]:
        pns = items[1]
        postcodes: List[str] = []

        # Is it a postcode?
        if re.match(r"\d\d\d$", pns[0]):
            postcodes.append(pns[0])
        else:
            # Try to look up placename
            pc = postcodes_for_placename(pns[0].strip(), partial=True)
            if pc:
                postcodes.extend([str(x) for x in pc])

        if postcodes:
            qp = " OR ".join([" postnr=? " for _ in postcodes])
            q += " AND (%s) " % qp
            qargs.extend(postcodes)

    q += " ORDER BY postnr ASC, husnr ASC, bokst ASC LIMIT ?"
    qargs.append(str(limit))

    return _run_addr_query(q, qargs)


def nearest_addr(lat: float, lon: float, limit: int = 1) -> List[Dict[str, Any]]:
    """Find the address closest to the given coordinates."""
    q = "SELECT * FROM stadfong"
    db_conn = shared_db.connection()
    res = db_conn.cursor().execute(q, [])
    closest = sorted(
        res, key=lambda i: distance((lat, lon), (i["lat_wgs84"], i["long_wgs84"]))
    )
    return [_add_postcode_info(x) for x in closest[:limit]]


# def format_addr(addr: Dict[str, Any]) -> str:
#     """Given an address record dict from the database,
#     returns a canonically formatted address string, e.g.
#     'Öldugata 9c, 101 Reykjavík'"""
#     pass
